import javax.tools.JavaCompiler;
import java.io.*;
import java.util.*;

import java.util.concurrent.*;

public class Main {

    //use stdin for test content (1 line)
    //TODO limit memory... currently we could use cgroups but problem with memswap (no permission and file not exists)
    //this would be static and could not be changed for every test run
    /*
    firejail can use cgroup

    sudo cgcreate -a appinventor -t appinventor -g memory:stest2
    echo 5000000 > /sys/fs/cgroup/memory/stest2/memory.limit_in_bytes //10 MB

    #set swapping to same value to not allow swapping
    #klappt nicht, nicht da, keine berechtigung
    echo 5000000 > /sys/fs/cgroup/memory/stest2/memory.memsw.limit_in_bytes
     */

    /**
     * the max number if lines to read in ONE WHILE loop in the extra output stream (after the program finished)
     * we use this because sometimes the stream.ready() function becomes never false and thus we have an infinite loop...
     * (probably that's the reason for the 100% infinitely running processes...)
     */
    static int MaxExtraOutputLinesToRead = 100;

    /**
     * the max number if lines to read in ONE WHILE loop in the error stream
     * we use this because sometimes the stream.ready() function becomes never false and thus we have an infinite loop...
     * (probably that's the reason for the 100% infinitely running processes...)
     */
    static int MaxErrorStreamLinesToRead = 100;

    /**
     * true: outputs debug information to stdout
     */
    static boolean isDebug = false;
    /**
     * true: use some debug test content to test,
     * false: use received test content
     */
    static boolean useDebugTestContent = false;

    /**
     * the compile test command
     */
    static String compileMethod = "compileTest";
    /**
     * the execute black box test command
     */
    static String executeBlackBoxTestMethod = "blackBoxTest";
    /**
     * the execute regex test command (not implemented should be done on the test server else we would need to write files to disk)
     */
    static String regexTestMethod = "regexTest";


    /**
     * the sign for special command parameters (compile and black box e.g. #1 = main file name)
     */
    static String SpecialParamsSign = "#";
    /**
     * the char that indicates the start of the command line arguments in the test content
     */
    static Character CommandLineArgumentChar = '$';
    /**
     * the char that indicates the input for the users program in the test content
     */
    static Character inputToProgramChar = '<';
    /**
     * the char that indicates the output for the users program in the test content
     */
    static Character outputFromProgramChar = '>';

    /**
     * the error (stream) prefix
     */
    static String errorOutputPrefix = "error:";

    /**
     * the expected output
     */
    static String expectedOutputPrefix = "expected:";

    /**
     * the string in the conversation that indicates the end of the test ... the program should not output something after this
     * END OF TEST
     */
    static String programEndString = "EOT";


    /**
     * the char that matches every user programs output EVENT NO OUTPUT! (this is used when the output does not matter)
     */
    static Character wildcardOutputMatchChar = '*';

    /**
     * the stored conversation / protocol of the test run
     */
    static ArrayList<String> Conversation = new ArrayList<>();
    /**
     * the arguments
     */
    static String[] args = null;
    /**
     * the rest result code
     */
    static int runResult = -1;
    /**
     * the directory path
     * used to not leak paths
     */
    static String DirectoryPath = "";

    //moved to the test server
    static String NewLineString = "\n";

    /**
     * the debug line to now allocate a new var every time
     */
    static String debugLine = null;

    /**
     * the profile name for firejail
     */
    static String FirejailProfilePath = "~/syndrom/syndrom.profile";
    /**
     * the cgroup name to use (currently not used)
     */
    static String DefaultCGroupName = "syndrom";


    public static void main(String[] myArgs) throws IOException {

        if (isDebug) {
            System.out.println(Arrays.toString(myArgs));
        }

        Main.args = myArgs;

        if (args.length == 5 && Objects.equals(args[0], compileMethod)) {

            DirectoryPath = args[1];

            runResult = compile(args[1], args[2], args[3], args[4]);

            System.exit(runResult);
            return;

        } else if (args.length == 10 && Objects.equals(args[0], executeBlackBoxTestMethod)) {


            DirectoryPath = args[1];
            runResult = executeTest(args);

            if (runResult != 100) { //100 = directory not exists ... some other error
                // for all other errors: program was partly executed ... write output
                OutputConversation();
            }

            System.exit(runResult);
            return;

        } else if (args.length == 3 && Objects.equals(args[0], regexTestMethod)) {
            //done in php (no need to compile or write any file
        }

        System.out.println("Unknown command, length of arguments: " + myArgs.length);
        System.exit(101);
        return;

    }

    /**
     * outputs the saved conversation to the stdout
     */
    public static void OutputConversation() {
        for (int i = 0; i < Conversation.size(); i++) {
            String line = Conversation.get(i);
            System.out.print(line + NewLineString);
        }
    }

    /**
     * compiles the given main file in the given directory with the given compile command
     *
     * @param directory                      the (absolute) directory path
     * @param mainFileNameWithExtension      the main file to use
     * @param compileCmd                     the compile command to use
     * @param sourceFileExtensionsWithoutDot the extensions of the soruce files for the current plang (to not include asset files in the compilation files)
     * @return the test result code
     */
    public static int compile(String directory, String mainFileNameWithExtension, String compileCmd, String sourceFileExtensionsWithoutDot) {

        File directoryFile = new File(directory);

        if (!directoryFile.isDirectory()) {
            System.out.println("Directory not found");
            return 100;
        }

        File mainFile = new File(directoryFile.getPath() + File.separator + mainFileNameWithExtension);

        if (!mainFile.exists()) {
            System.out.println("Main file not found");
            return 100;
        }

        //better use the tokenizer ... else we need to escape the arguments what is ... bad
        List<String> arguments = transformCommandWithSpecialStrings(directoryFile, mainFile, mainFileNameWithExtension, compileCmd, -1, -1, -1, sourceFileExtensionsWithoutDot);

        if (arguments == null) {
            System.out.println("Could not read other files");
            return 100;
        }

        if (isDebug) {
            System.out.println(Arrays.toString(arguments.toArray()));
        }

        AddSandboxCommands(arguments, -1, -1, -1);

        if (isDebug) {
            System.out.println("With sandbox: " + Arrays.toString(arguments.toArray()));
        }

        //better use arguments as list --> no need to escape arguments
        ProcessBuilder processBuilder = new ProcessBuilder(arguments);


        //ProcessBuilder processBuilder = new ProcessBuilder("C:\\Python27\\python.exe", "-m", "py_compile", "C:\\Users\\theju\\syndromTetsts\\1471724730\\main.py");

        //processBuilder.redirectErrorStream(true);
        processBuilder.directory(directoryFile);

        Process p = null;
        try {
            p = processBuilder.start();
        } catch (IOException e) {
            System.out.println("IOException compiling file: " + toSafeLine(e.getMessage()));
            return 100;
        }

        StreamGobbler errorGobbler = new StreamGobbler(p.getErrorStream());
        StreamGobbler outputGobbler = new StreamGobbler(p.getInputStream());
        errorGobbler.start();
        outputGobbler.start();

        int exitValue = 0;

        try {
            p.waitFor();

            exitValue = p.exitValue();

        } catch (InterruptedException e) {
            System.out.println("InterruptedException while compiling file");
            return 100;
        }

        try {
            outputGobbler.join();
            errorGobbler.join();
        } catch (IllegalMonitorStateException | InterruptedException e) { //when the process already
            //e.printStackTrace();
        }

        if (exitValue != 0) {
            //some (compile) error
            return 50;
        }

        return 0;
    }

    /**
     * executes a black box test with the given arguments
     *
     * @param args the arguments
     * @return the test result code
     */
    public static int executeTest(String[] args) {

        String directory = args[1];
        String mainFileNameWithExtension = args[2];
        String compileCmd = args[3];
        String executeTestCmd = args[4];
        String timeoutString = args[5];
        String memoryLimitString = args[6];
        String diskLimitString = args[7];
        String sourceFileExtensionsWithoutDot = args[8];
        String shouldCompileString = args[9];
        String arguments = "";

        boolean shouldCompile = true;

        //check if we need to compile or not
        if (shouldCompileString.equalsIgnoreCase("false")) {
            shouldCompile = false;
        }


        //now get the test content from stdin
        String testContentLine = "";
        ArrayList<String> testContents = new ArrayList<>();

        if (useDebugTestContent) {
            //testContents.add("$ 1");
            testContents.add("< 0");
            testContents.add("< 0");
            testContents.add("> 0");
            //testContents.add("> 1");
            //testContents.add("< 42");
            //testContents.add("> 42");
            //testContents.add("> 42");
            //testContents.add("< Test Echo");
            //testContents.add("> Test Echo");
        }


        Timer timer = new Timer();
        if (isDebug) {
            timer.start();
        }



        //read the test content from the stdin
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        try {
            if (in.ready()) {
                while ((testContentLine = in.readLine()) != null) {
                    testContents.add(testContentLine);
                }
            }
            in.close();
        } catch (IOException e) {
            System.out.println("Could not read test content (from stdin):");
            return 100;
        }

        if (isDebug) {
            System.out.println("test lines: " + Arrays.toString(testContents.toArray()));
        }


        int timeoutInMs = 0;
        int memoryLimitInKb = 0;
        int diskLimitInKb = 0;

        try {
            timeoutInMs = Integer.parseInt(timeoutString);
        } catch (NumberFormatException e) {
            System.out.println("Could not parse timeout");
            return 100;
        }

        try {
            memoryLimitInKb = Integer.parseInt(memoryLimitString);
        } catch (NumberFormatException e) {
            System.out.println("Could not parse memory limit");
            return 100;
        }

        try {
            diskLimitInKb = Integer.parseInt(diskLimitString);
        } catch (NumberFormatException e) {
            System.out.println("Could not parse disk limit");
            return 100;
        }

        File directoryFile = new File(directory);

        if (!directoryFile.isDirectory()) {
            System.out.println("Directory not found");
            return 100;
        }

        File mainFile = new File(directoryFile.getPath() + File.separator + mainFileNameWithExtension);

        if (!mainFile.exists()) {
            System.out.println("Main file not found");
            return 100;
        }

        if (isDebug) {
            timer.end();
            System.out.println("time before compile in ms :"  + timer.getTotalTimeInMs());
            timer.start();
        }

        //first compile... already done in php
        int compileResult = 0;

        //check if we need to compile
        if (shouldCompile) {
            compileResult = compile(directoryFile.getAbsolutePath(), mainFileNameWithExtension,
                    compileCmd, sourceFileExtensionsWithoutDot);
        }

        if (isDebug) {
            timer.end();
            System.out.println("time for compiling in ms :"  + timer.getTotalTimeInMs());
        }

        if (compileResult != 0) {
            //some error
            return compileResult;
        }



        final ArrayList<String> testLines = testContents; //testContent.split("\n");

        List<String> _arguments = null;
        for (int i = 0; i < testLines.size(); i++) {
            String testLine = testLines.get(i).trim();

            if (testLine.length() > 0) {
                //this is really a test line, no white space line

                if (testLine.charAt(0) == CommandLineArgumentChar) {
                    //this is the argument line...
                    arguments = testLine.substring(1);
                    _arguments = ArgumentTokenizer.tokenize(arguments);
                    Conversation.add(testLine);
                    break;
                }
            }
        }

        List<String> cmdToEx = transformCommandWithSpecialStrings(directoryFile, mainFile,
                mainFileNameWithExtension, executeTestCmd, timeoutInMs, memoryLimitInKb, diskLimitInKb,
                sourceFileExtensionsWithoutDot);

        if (cmdToEx == null) {
            System.out.println("Could not read other files");
            return 100;
        }

        //add console arguments to the command list
        if (_arguments != null) {
            cmdToEx.addAll(_arguments);
        }

        if (isDebug) {
            System.out.println(Arrays.toString(cmdToEx.toArray()));
        }

        AddSandboxCommands(cmdToEx, timeoutInMs, memoryLimitInKb, diskLimitInKb);

        if (isDebug) {
            System.out.println("With sandbox: " + Arrays.toString(cmdToEx.toArray()));
        }

        ProcessBuilder processBuilder = new ProcessBuilder(cmdToEx);

        //processBuilder.redirectErrorStream(true);
        processBuilder.directory(directoryFile); //
        Process p = null;

        if (isDebug) {
            timer.start();
        }

        try {
            p = processBuilder.start();
        } catch (IOException e) {
            e.printStackTrace();
            System.out.println("Process could not be started");
            return 100;
        }

        //MAYBE always include error stream??

        final Boolean[] outputMismatch = {false};

        final Process finalP = p;
        final BufferedReader[] reader = new BufferedReader[1];
        final BufferedWriter[] writer = new BufferedWriter[1];
        final BufferedReader[] errReader = new BufferedReader[1];
        Callable<Integer> run = new Callable<Integer>() {
            @Override
            public Integer call() throws Exception {


                //InputStream errorStream = p.getErrorStream();
                InputStream inputStream = finalP.getInputStream();
                OutputStream outputStream = finalP.getOutputStream();

                reader[0] = new BufferedReader(new InputStreamReader(inputStream));
                writer[0] = new BufferedWriter(new OutputStreamWriter(outputStream));
                errReader[0] = new BufferedReader(new InputStreamReader(finalP.getErrorStream()));
                boolean acceptDebug = true;
                String debugLine = null;
                String output = "";
                //writer.close();

                for (int i = 0; i < testLines.size(); i++) {
                    String testLine = testLines.get(i).trim();

                    if (testLine.length() > 0) {

                        if (testLine.charAt(0) == inputToProgramChar) {

                            //before writing to the program the user maybe wants to debug output...

                            //readDebugFromErrorStream(errReader, false);

                            Conversation.add(testLine.trim());
                            if (isDebug) {
                                System.out.println(testLine.trim());
                            }

                            //feed program with input

                            try {
                                //seems as we really need the new line because readLine will only read a line terminated
                                //by a new line (the users program)
                                String line = testLine.substring(1).trim() + System.getProperty("line.separator");
                                //outputStream.write(line.getBytes());
                                writer[0].write(line);
                                writer[0].flush();

                            } catch (IOException e) {
                                System.out.println("Could not write to the process");
                                return 100;
                            }

                            //after writing to the program user maybe wants debug output...
                            //readDebugFromErrorStream(errReader, false);

                        } else if (testLine.charAt(0) == outputFromProgramChar) {
                            //read the output (if any)


                            //before the real output is read maybe the user wants to debug output...
                            //readDebugFromErrorStream(errReader, false);

                            String tr = testLine.substring(1);
                            String trimmedTestLine = tr.trim();

                            try {

                                //a wildcard output ... output does not matter but we need output (TODO right??)
                                //if (trimmedTestLine.length() > 0 && trimmedTestLine.charAt(0) == wildcardOutputMatchChar) {
                                  //  if (reader[0].ready()) {
                                      //  output = reader[0].readLine();
                                    //}
                                //} else {
                                    //we expect output... try to read ... even if we block
                                    output = reader[0].readLine();
                                //}

                            } catch (IOException e) {
                                System.out.println("Could not read from the process");
                                //error prefix is only for the user program errors
                                //Conversation.add(errorOutputPrefix: ' could not read from the process');
                                return 100;
                            }

                            if (trimmedTestLine.length() > 0 && trimmedTestLine.charAt(0) == wildcardOutputMatchChar) {
                                //program output does not matter but we need output
                                Conversation.add(outputFromProgramChar + " " + toSafeLine(output.trim()));
                            } else {

                                String outPut = "";
                                //check if output matched
                                if (output == null) {
                                    //no output from user program
                                    //but we cannot output null...
                                } else {
                                    outPut = output;
                                }

                                if (Objects.equals(outPut.trim(), trimmedTestLine)) {

                                    Conversation.add(outputFromProgramChar + " " + trimmedTestLine); //force space
                                    if (isDebug) {
                                        System.out.println(outputFromProgramChar + " " + trimmedTestLine);
                                    }
                                } else {
                                    Conversation.add(outputFromProgramChar + " " + toSafeLine(outPut.trim()));
                                    Conversation.add(expectedOutputPrefix + " " + trimmedTestLine);
                                    if (isDebug) {
                                        System.out.println(outputFromProgramChar + " " + toSafeLine(outPut.trim()));
                                        System.out.print(expectedOutputPrefix + " " + trimmedTestLine);
                                    }
                                    outputMismatch[0] = true;
                                }
                            }

                            //after the users program output ... debug output?
                            //readDebugFromErrorStream(errReader, false);
                        } else {
                            //ignore line
                        }

                        //after every line check the stderr stream (this can be used to log debug messages)
                        //even this line stays here the order of the conversation might not be the same as the
                        //real output... e.g.
                        /*
                                System.out.println(0);
                                System.out.println(1);
                                System.err.println(test);

                                can result in
                                0
                                test
                                1

                                or
                                0
                                1
                                test

                                or any other order...

                                this might be because after the first output the stderr is
                                read but there is no error output
                                or something other

                         */
                        readDebugFromErrorStream(errReader, false);
                    }
                }

                Conversation.add(programEndString);
                /*
                try {
                    //reader[0].close(); //do not close reader here because we need to check if the program
                    //outputs something after the test (empty test --> program should output nothing)
                    //writer.close(); //closing will write (or not?) some special char?
                    //in.readLine() will throw a NumberFormatException exception : null
                    //same as scanner but with different error
                } catch (IOException e) {

                }
                */

                return finalP.waitFor();
                //return 0;
            }
        };

        ExecutorService executor = Executors.newSingleThreadExecutor();
        //Future<Integer> future = executor.submit(run);
        FutureTask<Integer> task = new FutureTask<Integer>(run);


        BufferedReader errorReader = null;
        int returnValue = 0;
        int futureResult = -100;
        int r = -100;

        try {

            executor.execute(task);

            futureResult = task.get(timeoutInMs, TimeUnit.MILLISECONDS);

            //furutreResult = future.get(timeoutInMs, TimeUnit.MILLISECONDS);
            //p.waitFor(timeoutInMs, TimeUnit.MILLISECONDS);
            //r = p.exitValue(); //this blocks the thread until process finished

            //} catch (TimeoutException | InterruptedException e) {
        } catch (InterruptedException e) {
            //timeout hit
            returnValue = 1;
            futureResult = -100;
        } catch (ExecutionException e) {
            //this also happens if the program (to test) throws an error

            //System.out.println("ExecutionException occurred");
            returnValue = 2;
            futureResult = -100;
        } catch (TimeoutException e) {
            //timeout hit
            returnValue =  3;
            futureResult = -100;
        } catch (IllegalThreadStateException e) {
            //executor.shutdownNow();
            p.destroy(); //TODO does not stop child processes --> http://stackoverflow.com/questions/6356340/killing-a-process-using-java
            returnValue = 3;
            futureResult = -100;
        }


        //p.waitFor will NOT TERMINATE the process in any way...
        //... so finally kill the process (if it's not already finished)
        //else an infinitely loop would run forever
        p.destroy();

        if (isDebug) {
            timer.end();
            System.out.println("time for running process in ms :"  + timer.getTotalTimeInMs());
        }

        //after the test has finished close streams
        try {
            //read debug before reading output that is too much
            readDebugFromErrorStream(errReader, true);

            String line = null;

            if (returnValue == 0) {
                //process has finished normally so we can safely check the reader.ready()
                //when there is a timeout (because user program waits for input  reader.ready() would block this program...
                //but the program terminated normally here...

                //TODO MAYBE or swap the if and while ??
                int count = 0;
                if (reader[0].ready()) { //this line blocks if the process has not yet ended (user program waits for input??)
                    //if the test ended and the program outputs something the test should fail...
                    while (((line = reader[0].readLine()) != null) && count < MaxExtraOutputLinesToRead ) {
                        Conversation.add(outputFromProgramChar + " " + toSafeLine(line.trim()));
                        Conversation.add(expectedOutputPrefix + " " + ""); //no more output expected... -> empty
                        outputMismatch[0] = true;
                    }
                }
            } else {
                //process has ended in an unusual way...
                //Conversation.add(outputFromProgramChar + " " + toSafeLine(line.trim()));
                //Conversation.add(expectedOutputPrefix + " " + ""); //no more output expected... -> empty
                //pretend false output
                //outputMismatch[0] = true;

                //some thread exception ... pretend timeout...
                returnValue = 3;
            }

            writer[0].close();
            reader[0].close();
        } catch (IOException e) {
            //e.printStackTrace();
        }

        // 0 = all went well (java code)
        //-100 is timeout --> returnValue is 3
        //this if would change this to 2
        if (futureResult != 0 && futureResult != -100) {

            //else some error?
            errorReader = new BufferedReader(new InputStreamReader(p.getErrorStream()));

            //only read error if really an error occurred
            //if (errorReader != null) {
                String errorLine = "";

                try {
                    int count = 0;
                    if (errorReader.ready()) {
                        while (((errorLine = errorReader.readLine()) != null) && count < MaxErrorStreamLinesToRead) {
                            Conversation.add(errorOutputPrefix + " " + toSafeLine(errorLine));
                        }
                    }
                } catch (IOException e) {
                    //e.printStackTrace();
                }
            //}

            //some error occurred during execution of the users program (e.g. numberformat exception....)
            returnValue = 2;
        }

        if (outputMismatch[0]) {
            returnValue = 1;
        }

        return returnValue;
    }

    /**
     * tries to read from the error stream to read debugging messages
     * @param errReader the error reader
     */
    private static void readDebugFromErrorStream(BufferedReader[] errReader, boolean useWhile) {

        //when we use while instead of if we could read more lines...
        //but then we would get lines that might outputted after some normal output
        //... e.g.
                            /*
                            System.err.println("test0");
                            System.out.println("0");
                            System.err.println("min: " + min);

                            would output both error lines because of the while...
                             */

        //but if we do not use while we can only read 1 line and there could be more outputted... so use while
        //or the error stream is not yet read and the normal output was faster ... so we would have 2x normal 1x error
        //even if the right order was 1x normal 1 error 1x normal

        try {

            if (useWhile) {
                int count = 0;
                if (errReader[0].ready()) {
                    while (((debugLine = errReader[0].readLine()) != null) && count < MaxErrorStreamLinesToRead) {
                        Conversation.add(errorOutputPrefix + " " + toSafeLine(debugLine));
                        count++;
                    }
                }
            } else {
                if (errReader[0].ready()) {
                    if ((debugLine = errReader[0].readLine()) != null) {
                        Conversation.add(errorOutputPrefix + " " + toSafeLine(debugLine));
                    }
                }
            }

        } catch (IOException e) {
        }
    }

    /**
     * transforms the commandToExecute and replaces the special strings withe it's values
     *
     * @param directoryFile                  the base directory where all files are located
     * @param mainFileNameWithExtension      the main file
     * @param commandToExecute               the command to execute
     * @param timeoutInMs                    the timeout in ms (use 0 or negative value of not set)
     * @param memoryLimitInKb                the memory limit in kb (use 0 or negative value of not set)
     * @param diskLimitInKb                  the disk limit in kb (use 0 or negative value of not set)
     * @param sourceFileExtensionsWithoutDot the extensions of the soruce files for the current plang (to not include asset files in the compilation files)
     * @return the transformed command or null if the files could not be read
     */
    private static List<String> transformCommandWithSpecialStrings(File directoryFile, File mainFile,
                                                                   String mainFileNameWithExtension,
                                                                   String commandToExecute, int timeoutInMs,
                                                                   int memoryLimitInKb, int diskLimitInKb,
                                                                   String sourceFileExtensionsWithoutDot) {

        final String[] sourceFileExtensions = sourceFileExtensionsWithoutDot.split(",");

        String mainFileName = mainFileNameWithExtension.substring(0, mainFileNameWithExtension.lastIndexOf("."));

        List<String> executionParts = ArgumentTokenizer.tokenize(commandToExecute);

        StringBuilder fileNameList = new StringBuilder();
        StringBuilder fileNameWithExtensionsList = new StringBuilder();
        StringBuilder fileNameWithAbsolutePathList = new StringBuilder();
        StringBuilder fileNameWithAbsolutePathWithExtensionList = new StringBuilder();

        if (commandToExecute.contains(Main.SpecialParamsSign + "6")
                || commandToExecute.contains(Main.SpecialParamsSign + "7")
                || commandToExecute.contains(Main.SpecialParamsSign + "8")
                || commandToExecute.contains(Main.SpecialParamsSign + "9")) {

            //TODO just with the right extensions
            File[] listOfFiles = directoryFile.listFiles(new FileFilter() {
                @Override
                public boolean accept(File pathname) {
                    String fileName = pathname.getName();
                    String fileExtensions = fileName.substring(fileName.lastIndexOf(".") + 1); //extension list is without dots
                    //filter for source files (do not include assets
                    for (int i = 0; i < sourceFileExtensions.length; i++) {
                        if (Objects.equals(sourceFileExtensions[i], fileExtensions)) {
                            return true;
                        }
                    }
                    return false;
                }
            });

            if (listOfFiles == null) {
                System.out.println("error getting other files");
                return null;
            }

            for (int i = 0; i < listOfFiles.length; i++) {
                File file = listOfFiles[i];
                if (file.isFile()) {

                    String fileNameWithExtension = file.getName();
                    String fileName = fileNameWithExtension.substring(0, fileNameWithExtension.lastIndexOf("."));
                    String absolutePathWithFileNameWithExtension = file.getAbsolutePath();

                    if (Objects.equals(fileNameWithExtension, mainFileNameWithExtension)) {
                        //ignore just the other files
                    } else {
                        //String fileName = fileNameWithExtension.substring()
                        //fileList.append()
                        fileNameList.append(wrapArgument(fileName));
                        fileNameWithExtensionsList.append(wrapArgument(fileNameWithExtension));
                        fileNameWithAbsolutePathList.append(wrapArgument(absolutePathWithFileNameWithExtension.substring(0, absolutePathWithFileNameWithExtension.lastIndexOf("."))));
                        fileNameWithAbsolutePathWithExtensionList.append(wrapArgument(absolutePathWithFileNameWithExtension));

                        //separation for other files
                        if (i != listOfFiles.length - 1) {
                            fileNameList.append(" ");
                            fileNameWithExtensionsList.append(" ");
                            fileNameWithAbsolutePathList.append(" ");
                            fileNameWithAbsolutePathWithExtensionList.append(" ");
                        }
                    }
                }
            }
        }

        if (isDebug && commandToExecute.contains(Main.SpecialParamsSign + "50")) {
            System.out.println("compile command contains " + Main.SpecialParamsSign + "50" + " but compile command do not use this, set to empty");
        }
        if (isDebug && commandToExecute.contains(Main.SpecialParamsSign + "51")) {
            System.out.println("compile command contains " + Main.SpecialParamsSign + "51" + " but compile command do not use this, set to empty");
        }
        if (isDebug && commandToExecute.contains(Main.SpecialParamsSign + "51")) {
            System.out.println("compile command contains " + Main.SpecialParamsSign + "52" + " but compile command do not use this, set to empty");
        }

        //inject special char values
        for (int i = 0; i < executionParts.size(); i++) {
            String executionPart = executionParts.get(i);

            if (executionPart.contains(Main.SpecialParamsSign + "1")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "1", (directoryFile.getAbsolutePath())));

            } else if (executionPart.contains(Main.SpecialParamsSign + "2")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "2", mainFileName));

            } else if (executionPart.contains(Main.SpecialParamsSign + "3")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "3", mainFileNameWithExtension));

            } else if (executionPart.contains(Main.SpecialParamsSign + "4")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "4", directoryFile.getAbsolutePath() + File.separator + mainFileName));

            } else if (executionPart.contains(Main.SpecialParamsSign + "5")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "5", mainFile.getAbsolutePath()));


            } else if (executionPart.contains(Main.SpecialParamsSign + "6")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "6", fileNameList.toString()));

            } else if (executionPart.contains(Main.SpecialParamsSign + "7")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "7", fileNameWithExtensionsList.toString()));

            } else if (executionPart.contains(Main.SpecialParamsSign + "8")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "8", fileNameWithAbsolutePathList.toString()));

            } else if (executionPart.contains(Main.SpecialParamsSign + "9")) {
                executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "9", fileNameWithAbsolutePathWithExtensionList.toString()));


            } else if (executionPart.contains(Main.SpecialParamsSign + "50")) {

                if (timeoutInMs <= 0) {
                    executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "50", timeoutInMs + ""));
                } else {
                    executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "50", ""));
                }

            } else if (executionPart.contains(Main.SpecialParamsSign + "51")) {
                if (memoryLimitInKb <= 0) {
                    executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "51", memoryLimitInKb + ""));
                } else {
                    executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "51", ""));
                }

            } else if (executionPart.contains(Main.SpecialParamsSign + "52")) {
                if (diskLimitInKb <= 0) {
                    executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "52", diskLimitInKb + ""));
                } else {
                    executionParts.set(i, executionPart.replace(Main.SpecialParamsSign + "52", ""));
                }

            }

        }

        return executionParts;
    }


    /**
     * adds some commands to the given commands to execute the given commands in a sandbox
     * @param commands the command as array
     * @param timeoutInMs the timeout to use
     * @param memoryLimitInKb the memory limit to use
     * @param diskLimitInKb the disk limit to use
     * @return  the commands extended with the commands to execute the sandbox
     */
    public static List<String> AddSandboxCommands(List<String> commands, int timeoutInMs, int memoryLimitInKb, int diskLimitInKb) {

        return commands;
    }

    /**
     * replaces every directory path (Main.DirectoryPath) found in the string (to not leak the executive directory)
     *
     * @param line the line
     * @return the line without the directory path
     */
    public static String toSafeLine(String line) {

        if (!Main.DirectoryPath.endsWith(File.separator)) {
            Main.DirectoryPath = Main.DirectoryPath + File.separator;
        }

        //when compiling 2 files A is the main file B not and B has an error then B the path of B is something like ./B.java
        if (line.startsWith("./") || line.startsWith(".\\")) {
            line = line.substring(2);
        }

        return line.replace(Main.DirectoryPath, "");
    }

    /**
     * somehow the arguments needs to be wrapped by "..." when executing
     *
     * @param line the line to wrap
     * @return the wrapped line e.g. "line"
     */
    public static String wrapArgument(String line) {
        return "\"" + line + "\"";
    }

}

//from http://stackoverflow.com/questions/1732455/redirect-process-output-to-stdout
class StreamGobbler extends Thread {
    InputStream is;

    // reads everything from is until empty.
    StreamGobbler(InputStream is) {
        this.is = is;
    }

    public void run() {
        try {
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);
            String line = null;
            String safeLine = null;
            while ((line = br.readLine()) != null) {
                safeLine = Main.toSafeLine(line);
                System.out.println(safeLine);
            }

        } catch (IOException ioe) {
            //ioe.printStackTrace();
        }
    }
}